# Mechanism Wizard 0.17 (21.05.2021)

from vcCommand import *
from math import *
import vcMatrix, vcVector

app = getApplication()
cmd = getCommand()
snapcmd = app.findCommand('interactiveSnap')


R2D = 57.296779513082320876798154814105
D2R = 0.017453292519943295769236907684886
EPSILON = 1e-6
AXIS_MAX_ANGLE_ERROR = 0.01
QUARILATERAL_ANGLE_TOLERANCE = 1.0
DOF_ZERO_TOLERANCE = 1.0


def OnStart():
  global comp, link1_prop, link2_prop, cp_prop, select_cp_prop, solve_prop
  
  context = app.AuthorContext
  comp = context.ActiveComponent
  if not comp:
    print 'No component selected.'
    return
  
  link1_prop = cmd.getProperty('Link1')
  link2_prop = cmd.getProperty('Link2')
  cp_prop = cmd.getProperty('Connection Point')
  select_cp_prop = cmd.getProperty('Select Connection Point')
  solve_prop = cmd.getProperty('Solve Follower Joints')
  
  if not link1_prop:
    link1_prop = cmd.createProperty(VC_STRING, 'Link1', VC_PROPERTY_STEP)
    link2_prop = cmd.createProperty(VC_STRING, 'Link2', VC_PROPERTY_STEP)
    cp_prop = cmd.createProperty(VC_VECTOR, 'Connection Point')
    #cp_prop.WritableWhenConnected = False
    #cp_prop.WritableWhenDisconnected = False
    #cp_prop.WritableWhenSimulating = False
    select_cp_prop = cmd.createProperty(VC_BUTTON, 'Select Connection Point')
    solve_prop = cmd.createProperty(VC_BUTTON, 'Solve Follower Joints')
  #endif
  
  # ATM always overrun old values to make sure that step values are up to date
  nodes = get_child_nodes(comp)
  node_names = [x.Name for x in nodes]
  link1_prop.StepValues = node_names
  link1_prop.Value = node_names[0]
  link2_prop.StepValues = node_names
  link2_prop.Value = node_names[0]
  select_cp_prop.OnChanged = select_connection_point
  solve_prop.OnChanged = solve
  
  app.clearMessages()
  
  print '--------------------'
  print 'Mechanism Wizard'
  print '--------------------'
  print 'Wizard is used to solve follower joints in closed-loop kinematics chains where you have 1 driving joint and 2 follower joints.'
  print 'Select two link chains that should be connected to form a closed loop. One chain should have 2 moving links and the other 1 moving link.'
  print 'Parameters:'
  print '-Link1: Link at the end of chain 1.'
  print '-Link2: Link at the end of chain 2.'
  print '-Connection Point: Connection point in component coordinates.'
  print '-Select Connection Point: Select connection point in 3D model.'
  print '-Solve Follower Joints: Solve follower joint expressions and write them in the links.'
  print '--------------------'
  
  
  executeInActionPanel()


def select_connection_point(*args):
  snapcmd.clear()
  snapcmd.Snap = True
  snapcmd.SnapOnCircleCenter = True
  snapcmd.SnapOnCurve = True
  snapcmd.SnapOnEdge = True
  snapcmd.SnapOnEdgeCenter = True
  snapcmd.SnapOnFaceCenter = True
  snapcmd.SnapOnFrame = True
  snapcmd.SnapOnNodeCenter = False
  snapcmd.SnapOnNodeOrigin = False
  snapcmd.SnapOnPoint = True
  snapcmd.SnapOnPolygon = True
  snapcmd.SnapOnSurface = True
  snapcmd.OnTargetSet = point_selected
  print 'Select connection point on model.'
  snapcmd.executeAsTool()


def point_selected(*args):
  snapcmd.OnTargetSet = None
  
  # Convert point to component coordiantes in case component is moved before solver
  p = comp.InverseWorldPositionMatrix * snapcmd.TargetPosition
  cp_prop.Value = p


def solve(*args):
  # Solve follower joint expressions
  
  link1 = comp.findNode(link1_prop.Value)
  link2 = comp.findNode(link2_prop.Value)
  connection_point = cp_prop.Value
  
  # Find out common parent node and both chains
  chain1 = get_parent_nodes(link1)
  chain2 = get_parent_nodes(link2)
  parent = None
  i = 0
  while i < len(chain1):
    link_i = chain1[i]
    j = 0
    while j < len(chain2):
      link_j = chain2[j]
      if link_i.Name == link_j.Name:
        parent = link_i
        break
      j += 1
    #endwhile
    if parent:
      break
    i += 1
  #endwhile
  
  # Strip chains up to parent node (don't include the parent node)
  chain1 = chain1[:i]
  chain2 = chain2[:j]
  if not parent or (len(chain1) + len(chain2) != 3):
    print 'Cannot solve system.'
    return
  
  # Find where driver is and mark that as chain1
  driver_in_chain1 = True
  for link in chain2:
    if link.JointType in [VC_JOINTTYPE_TRANSLATIONAL, VC_JOINTTYPE_ROTATIONAL]:
      driver_in_chain1 = False
      break
  if not driver_in_chain1:
    chainx = chain1
    chain1 = chain2
    chain2 = chainx
  
  # Get moving links into a single chain list (driver chain first)
  chainx = list(chain2)
  chainx.reverse()
  chain = list(chain1)
  chain.extend(chainx)
  
  #Find driver and followers (= system_links)
  driver_link = None
  driver_type = ''
  system_links = []
  for i in range(len(chain)):
    link = chain[i]
    if link.JointType in [VC_JOINTTYPE_TRANSLATIONAL, VC_JOINTTYPE_ROTATIONAL]:
      system_links = chain[i:] + chain[:i]
      driver_link = link
      if link.JointType == VC_JOINTTYPE_TRANSLATIONAL:
        driver_type = 'T'
      else:#if link.JointType == VC_JOINTTYPE_ROTATIONAL:
        driver_type = 'R'
  if not system_links:
    print 'Cannot solve system.'
    return
  follower_links = system_links[1:]
  follower_types = []
  for i in range(len(follower_links)):
    link = follower_links[i]
    if link.JointType in [VC_JOINTTYPE_ROTATIONAL_FOLLOWER, VC_JOINTTYPE_TRANSLATIONAL_FOLLOWER]:
      if link.JointType == VC_JOINTTYPE_TRANSLATIONAL_FOLLOWER:
        follower_types.append('T')
      else:#if link.JointType == VC_JOINTTYPE_ROTATIONAL_FOLLOWER:
        follower_types.append('R')
    elif link.JointType in [VC_JOINTTYPE_TRANSLATIONAL, VC_JOINTTYPE_ROTATIONAL]:
      if link.JointType == VC_JOINTTYPE_TRANSLATIONAL:
        driver_type = 'T'
      else:#if link.JointType == VC_JOINTTYPE_ROTATIONAL:
        driver_type = 'R'
  system_type = driver_type + ''.join(follower_types)
  if not driver_link or len(follower_links) != 2 or not system_type in ['RRR', 'RRT', 'RTR', 'TRR']:
    print 'Cannot solve system.'
    return
  
  if system_links[0].Dof.ValueExpression.replace(' ', '').upper() != 'VALUE':
    print 'Cannot solve system. Unsupported joint expression.'
    return
  
  # Find index where common node hits on system_links
  common_index = 1
  link = system_links[0]
  while link.Parent.Name != parent.Name:
    common_index += 1
    link = link.Parent
    
  # Find where connection point hits on system_links
  connection_index = common_index + len(chain2)
  
  #
  trans_direction = None
  rot_positions = []
  rot_directions = []
  for link in system_links:
    pos = comp.InverseWorldPositionMatrix * link.WorldPositionMatrix
    dir = None
    if link.Dof.AxisType == VC_JOINTAXIS_MINUS_X:
      dir = (-1.0) * pos.N
    elif link.Dof.AxisType == VC_JOINTAXIS_PLUS_X:
      dir = (1.0) * pos.N
    elif link.Dof.AxisType == VC_JOINTAXIS_MINUS_Y:
      dir = (-1.0) * pos.O
    elif link.Dof.AxisType == VC_JOINTAXIS_PLUS_Y:
      dir = (1.0) * pos.O
    elif link.Dof.AxisType == VC_JOINTAXIS_MINUS_Z:
      dir = (-1.0) * pos.A
    else:#if link.Dof.AxisType == VC_JOINTAXIS_PLUS_Z:
      dir = (1.0) * pos.A
    if link.JointType in [VC_JOINTTYPE_ROTATIONAL, VC_JOINTTYPE_ROTATIONAL_FOLLOWER]:
      rot_positions.append(pos.P)
      rot_directions.append(dir)
    if link.JointType in [VC_JOINTTYPE_TRANSLATIONAL, VC_JOINTTYPE_TRANSLATIONAL_FOLLOWER]:
      trans_direction = dir
  #endfor
  if len(rot_positions) == 2:
    # Non-RRR system, insert connection point as third rotational position and direction
    pos = vcVector.new(connection_point)
    dir = vcVector.new(rot_directions[0])
    index = connection_index
    if 'T' in system_type[:index]:
      index -= 1
    rot_positions.insert(index, pos)
    rot_directions.insert(index, dir)
  if not trans_direction:
    # RRR system, take direction from farther chain end and connection point
    chain_end_pos1 = comp.InverseWorldPositionMatrix * chain[0].WorldPositionMatrix
    chain_end_pos2 = comp.InverseWorldPositionMatrix * chain[-1].WorldPositionMatrix
    dir1 = connection_point - chain_end_pos1.P
    dir2 = connection_point - chain_end_pos2.P
    if dir1.length() > dir2.length():
      trans_direction = dir1
    else:
      trans_direction = dir2
  if len(rot_positions) != 3 or len(rot_directions) != 3:
    print 'Cannot solve system.'
    return
  
  # Check that rotational directions are parallel and form the 2D plane where to solve the problem
  for i in range(3):
    for j in range(i + 1, 3):
      rot_dir1 = rot_directions[i]
      rot_dir2 = rot_directions[j]
      angle = rot_dir1.angle(rot_dir2) * R2D
      if AXIS_MAX_ANGLE_ERROR < angle < 180.0 - AXIS_MAX_ANGLE_ERROR:
        print 'Cannot solve system. Rotational axes are not parallel.'
        return
  N = vcVector.new(rot_directions[0])
  P = vcVector.new(rot_positions[0])
  plane = vcMatrix.new()
  plane.N = N
  plane.P = P
  plane.uniform()
  iplane = vcMatrix.new(plane)
  iplane.invert()
  plane0 = vcMatrix.new(plane)
  plane0.P = vcVector.new()
  iplane0 = vcMatrix.new(plane0)
  iplane0.invert()
  
  # Convert problem to 2D plane, remove Z component and convert problem back to 3D (component origin)
  for i in range(3):
    rot_pos = rot_positions[i]
    rot_pos = iplane * rot_pos
    rot_pos.X = 0
    rot_pos = plane * rot_pos
    rot_positions[i] = rot_pos
  if system_type == 'RRR':
    # Do plane trick for trans_direction as it was calculated using snapped point so there might be unwanted 3rd component in it
    trans_direction = iplane0 * trans_direction
    trans_direction.X = 0
    trans_direction = plane0 * trans_direction
    trans_direction.normalize()
  
  # Check that translational joint is perpendicular to rotational joints
  angle = trans_direction.angle(plane.N) * R2D
  if abs(angle - 90.0) > AXIS_MAX_ANGLE_ERROR:
    print 'Cannot solve system. Translational axis is not perpendicular to rotational axes.'
    return
  
  #Solve triangle sides and angles
  side1 = rot_positions[1] - rot_positions[0]
  side2 = rot_positions[2] - rot_positions[1]
  side3 = rot_positions[0] - rot_positions[2] 
  angles = [180.0 - side3.angle(side1) * R2D, 180.0 - side1.angle(side2) * R2D, 180.0 - side2.angle(side3) * R2D]
  lengths = [side1.length(), side2.length(), side3.length()]
  
  #
  # MID RESULTS:
  #
  # parent              common node
  # chain1              contains selected link chain with driver link in backwards order                                                                                                           Example: [link2, link1]
  # chain2              contains selected link chain without driver link in backwards order                                                                                                        Example: [link3]
  # chain               contains moving links in order of: driver chain end => common node (not included) =>  follower chain end.                                                                  Example: [link2, link1, link3]
  # system_links        contains moving links in order of: driver link => common node (not included) => follower chain end => driver_chain_end => driver link (not included at the end)  e.g.      Example: [link1, link3, link2]
  # system_Type         joint types of system_links                                                                                                                                                Example: 'TRR'
  # common_index        index on system_links where common node point lands                                                                                                                        Example: 1
  # connection_index    index on system_links where connection point lands                                                                                                                         Example: 2
  # ~~~
  # rot_positions       positions of rotational joints (for non-RRR connection point pos is inserted in the list)
  # rot_directions      directions of rotational joints (for non-RRR connection point dir is inserted in the list)
  # trans_direction     direction of translational joint (for 'RRR' constructed based on connection point)
  # plane               2D plane formed at first rotational joint, plane.N is first rotational direction and plane.P is first rotational position
  # iplane              inverse of plane
  # plane0              same as plane but without translation (P is zero vector)
  # iplane0             inverse of plane0
  # angles              triangle corner angles
  # lengths             triangle side lengths
  #
  #
  #
  # EXAMPLE NODE TREE:
  #
  # parent
  # |
  # |--link1 (T, driver)
  # |  |
  # |  |--link2 (R, follower)
  # |
  # |--link3 (R, follower)
  #
  
  # System subtype 2, including parent node (p) and connection point (c)
  system_type_s2 = system_type
  system_type_s2 = system_type_s2[:connection_index] + 'c' + system_type_s2[connection_index:]
  system_type_s2 = system_type_s2[:common_index] + 'p' + system_type_s2[common_index:]
  
  #
  # TRIANGLE SOLVERS
  #
  # ok     permutation solved
  # trXi   rot indexes around trans
  # nrXs   rot natural signs
  # cri    index for connection point in rot_positions
  # fXe    follwer joint expression
  #
  ok, tr1i, tr2i, nr1s, nr2s, nr3s, cri, f1e, f2e = False, 0, 0, 1.0, 1.0, 1.0, 0, '', ''
  if system_type == 'RRR':
    if system_type_s2 == 'RpRcR':
      ok, tr1i, tr2i, nr1s, nr2s, nr3s, cri, f1e, f2e = True, 1, 2, -1.0, 1.0, -1.0, -1, 'sasad(S3, A1+VALUE, S1)-A2', 'sasad(S1, A1+VALUE, S3)-A3'
    elif system_type_s2 == 'RpRRc':
      pass
    elif system_type_s2 == 'RRpRc':
      pass
  elif system_type == 'TRR':
    if system_type_s2 == 'TpRcR':
      ok, tr1i, tr2i, nr1s, nr2s, nr3s, cri, f1e, f2e = True, 0, 2, 1.0, 1.0, -1.0, 1, 'sssad(S2, S1, S3+VALUE)-A1', 'sssad(S1, S2, S3+VALUE)-A3'
    elif system_type_s2 == 'TpRRc':
      ok, tr1i, tr2i, nr1s, nr2s, nr3s, cri, f1e, f2e = True, 0, 2, 1.0, 1.0, 1.0, 2, 'sssad(S2, S1, S3+VALUE)-A1', 'sssad(S3+VALUE, S2, S1)-A2'
    elif system_type_s2 == 'TRpRc':
      ok, tr1i, tr2i, nr1s, nr2s, nr3s, cri, f1e, f2e = True, 0, 2, -1.0, 1.0, 1.0, 2, 'sssad(S2, S1, S3+VALUE)-A1', 'sssad(S3+VALUE, S2, S1)-A2'
  elif system_type == 'RTR':
    if system_type_s2 == 'RpcTR':
      ok, tr1i, tr2i, nr1s, nr2s, nr3s, cri, f1e, f2e = True, 2, 1, -1.0, 1.0, -1.0, 1, 'sassd(S1, A1+VALUE, S3)-S2', 'sasad(S1, A1+VALUE, S3)-A3'
    elif system_type_s2 == 'RpTcR':
      pass
    elif system_type_s2 == 'RpTRc':
      pass
    elif system_type_s2 == 'RTpRc':
      pass
  else:#if system_type == 'RRT':
    if system_type_s2 == 'RpRcT':
      pass
    elif system_type_s2 == 'RpRTc':
      ok, tr1i, tr2i, nr1s, nr2s, nr3s, cri, f1e, f2e = True, 1, 2, -1.0, 1.0, 1.0, 2, 'sasad(S3, A1+VALUE, S1)-A2', 'sassd(S3, A1+VALUE, S1)-S2'
    elif system_type_s2 == 'RRpTc':
      ok, tr1i, tr2i, nr1s, nr2s, nr3s, cri, f1e, f2e = True, 1, 2, -1.0, -1.0, 1.0, 2, 'sasad(S3, A1+VALUE, S1)-A2', 'sassd(S3, A1+VALUE, S1)-S2'
  #endif
  if not ok:
    print 'Cannot solve system.'
    return
  
  
  # Get reference direction vector for translation direction
  tr1 = rot_positions[tr1i]
  tr2 = rot_positions[tr2i]
  trans_ref_direction = tr2 - tr1
  trans_ref_direction.normalize()
  
  # Check if system is quad
  is_quad = False
  trans_angle = trans_ref_direction.angle(trans_direction) * R2D
  if trans_angle > QUARILATERAL_ANGLE_TOLERANCE and trans_angle < 180.0 - QUARILATERAL_ANGLE_TOLERANCE:
    is_quad = True
  
  # Figure out translational sign (ts)
  ts = 1.0 - 2.0 * (trans_angle > 90.0)
  
  # Triangle's rotation direction in plane
  cross = side1 ^ side2
  clockwise = True
  if (cross.length() > 1e-3) and (cross.angle(plane.N) * R2D < 90.0):
    clockwise = False
  ps = -1.0 + 2.0 * (clockwise)
  
  # Find out rotational axis signs (rxs)
  ar1s = 1.0 - 2.0 * (rot_directions[0].angle(plane.N) * R2D > 90.0)
  ar2s = 1.0 - 2.0 * (rot_directions[1].angle(plane.N) * R2D > 90.0)
  ar3s = 1.0 - 2.0 * (rot_directions[2].angle(plane.N) * R2D > 90.0)
  r1s = ps * ar1s * nr1s
  r2s = ps * ar2s * nr2s
  r3s = ps * ar3s * nr3s
  
  #
  # QUAD SOLVERS
  #
  if is_quad:
    
    # Find free corner
    if tr1i == 0 and tr2i == 1:
      nontri = 2
    elif tr1i == 0 and tr2i == 2:
      nontri = 1
    else:#if tr1i == 1 and tr2i == 2:
      nontri = 0
    nontr = rot_positions[nontri]
    
    # Check where 4th corner (r4) is formed
    plane1 = vcMatrix.new()
    plane1.A = trans_direction
    plane1.P = tr1
    uniformZ(plane1)
    plane2 = vcMatrix.new(plane1)
    plane2.P = tr2
    r41 = projectPointToPlane(tr1, plane2)
    r42 = projectPointToPlane(tr2, plane1)
    v0 = nontr - tr1
    v1 = r41 - tr1
    v2 = r42 - tr1
    cross0 = trans_ref_direction ^ v0
    cross1 = trans_ref_direction ^ v1
    cross2 = trans_ref_direction ^ v2
    r4_from_tr1 = cross0.angle(cross1) > cross0.angle(cross2)
    if r4_from_tr1:
      r4 = r41
    else:
      r4 = r42
      
    # Insert r4 to rot_positions
    r4i = 0
    if tr1i == 0 and tr2i == 1:
      r4i = 1
    elif tr1i == 0 and tr2i == 2:
      r4i = 3
    else:#if tr1i == 1 and tr2i == 2:
      r4i = 2
    rot_positions.insert(r4i, r4)
    
    #Solve quad
    side1 = rot_positions[1] - rot_positions[0]
    side2 = rot_positions[2] - rot_positions[1]
    side3 = rot_positions[3] - rot_positions[2] 
    side4 = rot_positions[0] - rot_positions[3] 
    angles = [180.0 - side4.angle(side1) * R2D, 180.0 - side1.angle(side2) * R2D, 180.0 - side2.angle(side3) * R2D, 180.0 - side3.angle(side4) * R2D]
    lengths = [side1.length(), side2.length(), side3.length(), side4.length()]
    
    ok, f1e, f2e = False, '', ''
    if system_type == 'RRR':
      if system_type_s2 == 'RpRcR':
        pass
      elif system_type_s2 == 'RpRRc':
        pass
      elif system_type_s2 == 'RRpRc':
        pass
    elif system_type == 'TRR':
      if system_type_s2 == 'TpRcR':
        if r4_from_tr1:
          ok, f1e, f2e = True, 'sasssa1d(S3, A4, S4+VALUE, S1, S2)-A1', 'sasssa1d(S4+VALUE, A4, S3, S2, S1)-A3'
        else:
          ok, f1e, f2e = True, 'sasssa1d(S3+VALUE, A4, S4, S1, S2)-A1', 'sasssa1d(S4, A4, S3+VALUE, S2, S1)-A3'
      elif system_type_s2 == 'TpRRc':
        pass
      elif system_type_s2 == 'TRpRc':
        if r4_from_tr1:
          ok, f1e, f2e = True, 'sasssa1d(S3, A4, S4+VALUE, S1, S2)-A1', 'sasssa2d(S3, A4, S4+VALUE, S1, S2)-A2'
        else:
          ok, f1e, f2e = True, 'sasssa1d(S3+VALUE, A4, S4, S1, S2)-A1', 'sasssa2d(S3+VALUE, A4, S4, S1, S2)-A2'
    elif system_type == 'RTR':
      if system_type_s2 == 'RpTcR':
        pass
      elif system_type_s2 == 'RpTRc':
        pass
      elif system_type_s2 == 'RTpRc':
        pass
    else:#if system_type == 'RRT':
      if system_type_s2 == 'RpRcT':
        pass
      elif system_type_s2 == 'RpRTc':
        pass
      elif system_type_s2 == 'RRpTc':
        pass
    #endif
    if not ok:
      print 'Cannot solve system because there is offset angle in translational joint.'
      return
  #endif is_quad:
  
  # Build expressions
  es = [f1e, f2e]
  replacements = [['S1', '%.3f' % lengths[0]], ['S2', '%.3f' % lengths[1]], ['S3', '%.3f' % lengths[2]], ['A1', '%.3f' % angles[0]], ['A2', '%.3f' % angles[1]], ['A3', '%.3f' % angles[2]]]
  if is_quad:
    replacements.extend([['S4', '%.3f' % lengths[3]], ['A4', '%.3f' % angles[3]]])
  for i in range(len(es)):
    e = es[i]
    for [old, new] in replacements:
      e = e.replace(old, new)
    es[i] = e
  [f1e, f2e] = es
  
  #
  # System signs and expressions
  #
  # ds    driver sign
  # fXs   follower sign
  # fXe   follower expression
  #
  if system_type == 'RRR':
    ds, f1s, f2s = r1s, r2s, r3s
  elif system_type == 'TRR':
    ds = ts
    if cri == 0:
      f1s, f2s = r2s, r3s
    elif cri == 1:
      f1s, f2s = r1s, r3s
    else:
      f1s, f2s = r1s, r2s
  elif system_type == 'RTR':
    ds = r1s
    f1s = ts
    if cri < 2:
      f2s = r3s
    else:
      f2s = r2s
  else:#if system_type == 'RRT':
    ds = r1s
    if cri < 2:
      f1s = r3s
    else:
      f1s = r2s
    f2s = ts
  
  # Sign replacements
  if ds < 0:
    f1e = f1e.replace('+VALUE', '-VALUE')
    f2e = f2e.replace('+VALUE', '-VALUE')
  if f1s < 0:
    f1e = '-(%s)' % f1e
  if f2s < 0:
    f2e = '-(%s)' % f2e
  
  # Add initial position effect
  # Driver
  if abs(system_links[0].Dof.VALUE) > DOF_ZERO_TOLERANCE:
    init_exp = '%.3f' % (-system_links[0].Dof.VALUE)
    if not '-' in init_exp:
      init_exp = '+' + init_exp
    f1e = f1e.replace('VALUE', '(VALUE%s)' % (init_exp))
    f2e = f2e.replace('VALUE', '(VALUE%s)' % (init_exp))
  # Follower 1
  init_exp = system_links[1].Dof.ValueExpression.replace('VALUE', '%.3f' % (system_links[1].Dof.VALUE))
  try:
    init = eval(init_exp)
    if abs(init) > DOF_ZERO_TOLERANCE:
      init_exp = '%.3f' % (init)
      if not '-' in init_exp:
        init_exp = '+' + init_exp
      f1e = f1e + init_exp
  except:
    print 'Cannot solve system. Unsupported joint expression.'
    return
  # Follower 2
  init_exp = system_links[2].Dof.ValueExpression.replace('VALUE', '%.3f' % (system_links[2].Dof.VALUE))
  try:
    init = eval(init_exp)
    if abs(init) > DOF_ZERO_TOLERANCE:
      init_exp = '%.3f' % (init)
      if not '-' in init_exp:
        init_exp = '+' + init_exp
      f2e = f2e + init_exp
  except:
    print 'Cannot solve system. Unsupported joint expression.'
    return
  
  # Solve followers
  system_links[1].Dof.ValueExpression = f1e
  system_links[2].Dof.ValueExpression = f2e
  tri_quad = 'Triangle'
  if is_quad:
    tri_quad = 'Quadrilateral'
  
  print 'System type: %s (%s)' % (system_type, tri_quad)
  print 'Follower 1, %s: %s' % (system_links[1].Name, f1e)
  print 'Follower 2, %s: %s' % (system_links[2].Name, f2e)
  
  
  return


def get_child_nodes(node):
  nodes = [node]
  for child in node.Children:
    if child.Component == node.Component:
      nodes.extend(get_child_nodes(child))
  return nodes


def get_parent_nodes(node):
  nodes = [node]
  while node.Parent and node.Parent.Component == node.Component:
    node = node.Parent
    nodes.append(node)
  return nodes


def uniformZ(m):
  #Uniform matrix m so that Z axis (A) retains its direction. Second most important axis is X (N).
  #Modifies original matrix, return nothing.
  if not m:
    return
  a = m.A
  a.normalize()
  if a.length() < EPSILON:
    a = vcVector.new(0, 0, 1)
  o = a ^ m.N
  if o.length() < EPSILON:
    n = vcVector.new(1, 0, 0)
    o = a ^ n
    if o.length() < EPSILON:
      n = vcVector.new(0, 1, 0)
      o = a ^ n
  o.normalize()
  n = o ^ a
  n.normalize()
  m.N = n
  m.O = o
  m.A = a


def projectPointToPlane(p, m, a = None):
  #Project point p (vector) to plane m (matrix, m.A is plane normal).
  #If vector a is given, projection is done with that directional vetcor, 
  #otherwise plane normal m.A is used.
  if a:
    a.normalize()
    m.A = a
  else:
    uniformZ(m)
  im = vcMatrix.new(m)
  im.invert()
  p = im * p
  p.Z = 0
  p = m * p
  return p


##
## Triangular -solvers
##
def sasa( side1, angle3, side2 ):
  '''return angle between side2 and side3'''
  s1 = float(side1)
  a3 = float(angle3)
  s2 = float(side2)
  if s1 <= 0.0:
    return 0.0
  if s2 <= 0.0:
    return atan2(sin(a3),-cos(a3))
  s1_2 = s1*s1
  s2_2 = s2*s2
  s3_2 = s1_2 + s2_2 - 2*s1*s2*cos(a3)
  if s3_2 <= 0.0:
    return PI
  s3 = sqrt( s3_2 )
  cos_a1 = (s2_2 + s3_2 - s1_2)/(2*s2*s3)
  return atan2( s1*sin(a3)/s3, cos_a1 )

def sass( side1, angle3, side2 ):
  '''return side opposite angle3'''
  s1 = float(side1)
  a3 = float(angle3)
  s2 = float(side2)
  if s1 <= 0.0:
    if s2 < 0.0:
      return 0.0
    else:
      return s2
  if s2 <= 0.0:
    return s1
  s3 = s1*s1 + s2*s2 - 2*s1*s2*cos(a3)
  if s3 <= 0.0: 
    return 0.0
  return sqrt( s3 )

def sssa( side1, side2, side3 ):
  '''return angle between side2 and side3'''
  s1 = float(side1)
  s2 = float(side2)
  s3 = float(side3)
  if s1 <= 0.0 or s3 <= 0.0: 
    return 0.0
  if s2 <= 0.0:
    return PI/2.
  cos_a1 = (s2*s2 + s3*s3 - s1*s1)/(2*s2*s3)
  if cos_a1 > 1.0:
    cos_a1 = 1.0
  elif cos_a1 < -1.0:
    cos_a1 = -1.0
  return abs(acos( cos_a1 ))

# Degree version of same
def sasad( side1, angle3, side2 ):
  return sasa( side1, angle3*D2R, side2 )*R2D

def sassd( side1, angle3, side2 ):
  return sass( side1, angle3*D2R, side2 )

def sssad( side1, side2, side3 ):
  return sssa( side1, side2, side3 )*R2D


##
## Quadrilateral -solvers
##
def sasasa(s1,a1,s2,a2,s3): 
  ''''return angle between s3 and s4 (angles in radians)'''
  a1 = float(a1)
  a2 = float(a2)
  s1 = float(s1)
  s2 = float(s2)
  s3 = float(s3)
  return sasa( sass(s1,a1,s2), a2-sasa(s1,a1,s2), s3)

def sasass(s1,a1,s2,a2,s3): 
  '''return s4 (angles in radians)'''
  a1 = float(a1)
  a2 = float(a2)
  s1 = float(s1)
  s2 = float(s2)
  s3 = float(s3)
  return sass( sass(s1,a1,s2), a2-sasa(s1,a1,s2), s3)

def sasssa1(s1,a1,s2,s3,s4): 
  '''return angle between s2 and s3 (angles in radians)'''
  a1 = float(a1)
  s1 = float(s1)
  s2 = float(s2)
  s3 = float(s3)
  s4 = float(s4)
  s5 = sass(s1,a1,s2)
  return sssa(s1,s2,s5) + sssa(s4,s3,s5)

def sasssa2(s1,a1,s2,s3,s4): 
  '''return angle between s3 and s4 (angles in radians)'''
  a1 = float(a1)
  s1 = float(s1)
  s2 = float(s2)
  s3 = float(s3)
  s4 = float(s4)
  return sssa( sass(s1,a1,s2), s3, s4 )

# Degree version of same
def sasasad(s1,a1,s2,a2,s3): 
  ''''return angle between s3 and s4 (angles in degrees)'''
  a1 = float(a1)
  a2 = float(a2)
  s1 = float(s1)
  s2 = float(s2)
  s3 = float(s3)
  return sasasa(s1,a1*D2R,s2,a2*D2R,s3)*R2D

def sasassd(s1,a1,s2,a2,s3): 
  '''return s4 (angles in degrees)'''
  a1 = float(a1)
  a2 = float(a2)
  s1 = float(s1)
  s2 = float(s2)
  s3 = float(s3)
  return sasass(s1,a1*D2R,s2,a2*D2R,s3) 

def sasssa1d(s1,a1,s2,s3,s4): 
  '''return angle between s2 and s3 (angles in degrees)'''
  a1 = float(a1)
  s1 = float(s1)
  s2 = float(s2)
  s3 = float(s3)
  s4 = float(s4)
  return sasssa1(s1,a1*D2R,s2,s3,s4)*R2D

def sasssa2d(s1,a1,s2,s3,s4):
  '''return angle between s3 and s4 (angles in degrees)'''
  a1 = float(a1)
  s1 = float(s1)
  s2 = float(s2)
  s3 = float(s3)
  s4 = float(s4)
  return sasssa2(s1,a1*D2R,s2,s3,s4)*R2D